from _typeshed import Incomplete
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from queue import Queue
from typing_extensions import Optional

class Scheduler:
    lock: Incomplete
    queue: Queue[Optional[AbstractRequest]]
    seen_urls: set[str]
    max_retry_count: Incomplete
    def __init__(self, max_retry_count: int = 3) -> None:
        self.__no_more_data_flag = None
        self.__retry_dict = None
        ...
    def seen(self, url: str) -> bool: ...
    def push(self, req: Optional[AbstractRequest]) -> None: ...
    def pop(self) -> Optional[AbstractRequest]: ...
    def push_retry(self, req: AbstractRequest) -> bool: ...
    def set_no_more_data_flag(self) -> None: ...
    def get_no_more_data_flag(self) -> bool: ...



import threading
from _typeshed import Incomplete
from m_crawler.engine.abstract_engine import AbstractEngine as AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from m_crawler.pipeline.abstracy_pipeline import AbstractPipeline as AbstractPipeline
from m_crawler.pipeline.json_pipline import JsonPipeline as JsonPipeline
from m_crawler.scheduler.scheduler import Scheduler as Scheduler
from m_crawler.utils.count_down import CountDown as CountDown
from m_crawler.utils.crawler_exception import MaxRetryTimeException as MaxRetryTimeException

class Crawler(threading.Thread):
    lock: Incomplete
    engine: Incomplete
    scheduler: Incomplete
    pipeline: Incomplete
    threads: Incomplete
    count_down: CountDown
    def __init__(self, engine: AbstractEngine, pipeline: AbstractPipeline = ..., scheduler: Scheduler = ..., threads: int = 1) -> None: 
        self.__futures = None
        ...
    def append_request(self, *requests: AbstractRequest | None) -> None: ...
    def run(self) -> None: ...




from typing import Optional
from playwright.sync_api import sync_playwright
from playwright.sync_api import BrowserContext, Playwright, Browser
from m_crawler.config import PlayWeightConfig
from loguru import logger
from m_crawler.engine.abstract_engine import AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.engine.request import PlayWeightRequest
from m_crawler.engine.response import BaseResponse


class PlayWeightEngine(AbstractEngine):
    __slots__ = ("__playwright", "browser", "browser_context", "settings")
    """
    下载引擎
    """

    def __init__(self, settings: PlayWeightConfig) -> None:
        self.settings = settings
        self.__playwright: Optional[Playwright] = None
        self.browser: Optional[Browser] = None

        self.browser_context: Optional[BrowserContext] = None

    def __init_context(self) -> None:
        self.__playwright = sync_playwright().start()
        self.browser = self.__playwright.chromium.launch(
            executable_path=self.settings.executable_path,
            channel=self.settings.channel,
            args=self.settings.args,
            timeout=self.settings.timeout,
            env=self.settings.env,
            headless=self.settings.headless,
            downloads_path=self.settings.downloads_path,
            slow_mo=self.settings.slow_mo,
        )
        self.browser_context = self.browser.new_context()

    def download(self, request: AbstractRequest) -> AbstractResponse:
        """
        playweight 核心逻辑
        """
        # 确保传入的是 PlayWeightRequest 类型
        if not isinstance(request, PlayWeightRequest):
            raise TypeError("request must be an instance of PlayWeightRequest")
        self.__init_context()
        if not self.browser_context:
            raise ValueError("browser_context is not initialized")
        try:
            new_page = self.browser_context.new_page()
            new_page.goto(request.url)
            for action in request.actions:
                action(new_page)
            return BaseResponse(request, new_page.content(), True)
        except Exception as e:
            logger.error(f"URL:{request.url} failed, error:{e}")
            return BaseResponse(request, None, False)

    def after_download_hook(self, result: AbstractResponse) -> AbstractResponse:
        self.__browser_close()
        return result

    def __browser_close(self):
        if self.browser_context:
            self.browser_context.close()
        if self.browser:
            self.browser.close()



from concurrent.futures import Future, ThreadPoolExecutor, as_completed
import threading
from typing import Any, Optional

from loguru import logger
from m_crawler.engine.abstract_engine import AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.pipeline.abstracy_pipeline import AbstractPipeline
from m_crawler.pipeline.json_pipline import JsonPipeline
from m_crawler.scheduler.scheduler import Scheduler
from m_crawler.utils.count_down import CountDown
from m_crawler.utils.crawler_exception import MaxRetryTimeException


class Crawler(threading.Thread):

    lock = threading.Lock()

    def __init__(
        self,
        engine: AbstractEngine,
        pipeline: AbstractPipeline = JsonPipeline(),
        scheduler: Scheduler = Scheduler(),
        threads: int = 1,
        name: str = "Crawler",
        daemon: bool = True
    ) -> None:
        super().__init__(name=name, daemon=daemon)
        self.engine = engine
        self.scheduler = scheduler
        self.pipeline = pipeline
        self.threads = threads
        self.count_down: CountDown = CountDown()
        self.__futures:list[Future] = []

    def append_request(self, *requests: Optional[AbstractRequest]) -> None:
        for request in requests:
            self.scheduler.push(request)

    def run(self) -> None:
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            submit_thread = threading.Thread(target=self.__submit, args=(executor,))
            submit_thread.start()
            self.__watch()
            submit_thread.join()



    def __submit(self, executor: ThreadPoolExecutor) -> None:
        while True:
            request = self.scheduler.pop()
            if not request:
                if self.scheduler.get_no_more_data_flag():
                    break
                continue
            self.count_down.add()
            self.__futures.append(executor.submit(self.__run, request))

    def __run(self, request: AbstractRequest) -> Any:
        try:
            response: AbstractResponse = self.engine.run(request)
            if not response.success:
                raise RuntimeError()
            self.pipeline.process(response)
            return request.url
        except Exception as e:
            retriable = self.scheduler.push_retry(request)
            if not retriable:
                logger.error(e)
                raise MaxRetryTimeException()
            return None

    def __watch(self):
        for future in as_completed(self.__futures):
            try:
                result = future.result()
                if result:
                    logger.info(f"{result} completed")
            except Exception as e:
                logger.error(e)
            finally:
                self.count_down.down()
                if self.count_down.count_down() and self.scheduler.queue.empty():
                    self.scheduler.push(None)


from queue import Empty, Queue
import threading
from loguru import logger
from typing_extensions import Optional

from m_crawler.engine.abstract_request import AbstractRequest


class Scheduler:
    __slots__ = (
        "queue",
        "seen_urls",
        "max_retry_count",
        "__no_more_data_flag",
        "__retry_dict",
    )
    lock = threading.Lock()

    def __init__(self, max_retry_count: int = 3) -> None:
        self.queue: Queue[Optional[AbstractRequest]] = Queue()
        self.seen_urls: set[str] = set()
        self.max_retry_count = max_retry_count
        self.__retry_dict: dict[str, int] = {}
        self.__no_more_data_flag: bool = False

    def seen(self, url: str) -> bool:
        if url in self.seen_urls:
            return True
        with Scheduler.lock:
            if url in self.seen_urls:
                return True
            self.seen_urls.add(url)
            return False

    def push(self, req: Optional[AbstractRequest]) -> None:
        if req is None:
            self.queue.put(None)
        if req and not self.seen(req.url):
            self.queue.put(req)
            self.seen_urls.add(req.url)

    def pop(self) -> Optional[AbstractRequest]:
        try:
            return self.queue.get(block=True, timeout=30.0)
        except Empty:
            logger.error("Scheduler timeout")
            return None

    def push_retry(self, req: AbstractRequest) -> bool:
        # 检测是否已经达到重试的最大次数
        with Scheduler.lock:
            if self.__retry_dict.get(req.url, 0) >= self.max_retry_count:
                return False
            self.seen_urls.remove(req.url)
            self.__retry_dict[req.url] = self.__retry_dict.get(req.url, 0) + 1
            self.queue.put(req)
            logger.warning(f"Retrying {req.url}, Current retry count: {self.__retry_dict[req.url]}")
            return True

    def set_no_more_data_flag(self) -> None:
        """
        标识当前调度器外部的数据已经接收完毕，需要外部数据结束后调用此方法通知
        """
        self.__no_more_data_flag = True
        self.queue.put(None)

    def get_no_more_data_flag(self) -> bool:
        return self.__no_more_data_flag



import json
from typing import Any
from m_crawler.pipeline.abstracy_pipeline import AbstractPipeline


class JsonPipeline(AbstractPipeline):

    def process_item(self, item: Any) -> Any:
        json_item = json.dumps(item, ensure_ascii=False)
        print(json_item)



from abc import ABC, abstractmethod
from typing import Iterable, Optional, Set

from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.scheduler.scheduler import Scheduler


class AbstractPipeline(ABC):

    def __init__(self) -> None:
        self.schedulers: Optional[Set[Scheduler]] = None

    def register_scheduler(self, scheduler: Scheduler) -> None:
        if self.schedulers is None:
            self.schedulers = set()
        self.schedulers.add(scheduler)

    def process(self, item: Optional[AbstractResponse]) -> None:
        process_result = self.process_item(item)
        if self.schedulers:
            for scheduler in self.schedulers:
                for request in process_result:
                    if request:
                        scheduler.push(request)

    @abstractmethod
    def process_item(
        self, item: Optional[AbstractResponse]
    ) -> Iterable[Optional[AbstractRequest]]:
        """
        处理 engine 获取的数据
        """
        pass



import abc
from abc import ABC, abstractmethod
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from m_crawler.scheduler.scheduler import Scheduler as Scheduler
from typing import Iterable

class AbstractPipeline(ABC, metaclass=abc.ABCMeta):
    schedulers: set[Scheduler] | None
    def __init__(self) -> None: ...
    def register_scheduler(self, scheduler: Scheduler) -> None: ...
    def process(self, item: AbstractResponse | None) -> None: ...
    @abstractmethod
    def process_item(self, item: AbstractResponse | None) -> Iterable[AbstractRequest | None]: ...
